#include <gecode/int.hh>
#include <gecode/search.hh>
#include <gecode/minimodel.hh>
#include <gecode/support/timer.hh>

using namespace Gecode;

class Graceful : public Space {
protected:
	IntVarArray l;
public:
	Graceful(int n, int con) : l(*this, 2*n+1, 0, 4*n) {
		switch (con) {
			case 0:
				distinct(*this, l, ICL_VAL);
				break;
			case 1:
				distinct(*this, l, ICL_BND);
				break;
			case 2:
				distinct(*this, l, ICL_DOM);
				break;
		}
		IntVarArray d(*this, 4*n, 1, 4*n);
		int it = 0;
		for (int i = 0; i+1 < n; i++) {
				post(*this, d[it++] == abs(*this, minus(*this, l[i], l[i+1])));
				post(*this, d[it++] == abs(*this, minus(*this, l[n+i], l[n+i+1])));
		}
		for (int i = 0; i < n; i++) {
				post(*this, d[it++] == abs(*this, minus(*this, l[i], l[2*n])));
				post(*this, d[it++] == abs(*this, minus(*this, l[n+i], l[2*n])));
		}
				post(*this, d[it++] == abs(*this, minus(*this, l[0], l[n-1])));
				post(*this, d[it++] == abs(*this, minus(*this, l[n], l[2*n-1])));
		
		switch (con) {
			case 0:
				distinct(*this, d, ICL_VAL);
				break;
			case 1:
				distinct(*this, d, ICL_BND);
				break;
			case 2:
				distinct(*this, d, ICL_DOM);
				break;
		};
		for (int i = 0; i < 4*n; i++)
			l.add(*this, d[i]);

		branch(*this, l, INT_VAR_SIZE_MIN, INT_VAL_SPLIT_MAX);
	}

	Graceful(bool share, Graceful& s) : Space(share, s) {
		l.update(*this, share, s.l);
	}
	virtual Space* copy(bool share) {
		return new Graceful(share,*this);
	}

	void print(void) const {
		std::cout << l << std::endl;
	}
};

int main(int argc, char** argv) {
	if (argc != 3) {
		std::cout << "Usage: graceful_wn.gecode <consistency> <n>\n";
		return -1;
	}
	Support::Timer t;
	t.start();
	Graceful* model = new Graceful(atoi(argv[2]), atoi(argv[1]));
	DFS<Graceful> solver(model);
	if (Graceful* result = solver.next()) {
		std::cout << "SATISFIABLE\n";
		delete result;
	} else {
		std::cout << "UNSATISFIABLE\n";
	}
	Search::Statistics stats = solver.statistics();
	std::cout << "Time     \t: " << t.stop()/1000 << "\n";
	std::cout << "Conflicts\t: " << stats.fail << "\n";

	return 0;
}

